"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from test.unit.rules import BaseRuleTestCase
from test.unit.rules.templates.test_limitsize_template import write_limit_test_templates

from cfnlint.rules.templates.LimitDescription import (
    LimitDescription,  # pylint: disable=E0401
)


class TestTemplateLimitDescription(BaseRuleTestCase):
    """Test template limit size"""

    def setUp(self):
        """Setup"""
        super(TestTemplateLimitDescription, self).setUp()
        self.collection.register(LimitDescription())
        write_limit_test_templates()

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative(self):
        """Test failure"""
        self.helper_file_negative("test/fixtures/templates/bad/limit_size.yaml", 1)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import json
from test.testlib.testcase import BaseTestCase

import cfnlint.helpers
from cfnlint.rules import RulesCollection
from cfnlint.rules.resources.Configuration import Configuration  # pylint: disable=E0401
from cfnlint.runner import Runner


class TestInclude(BaseTestCase):
    """Used for Testing Rules"""

    def setUp(self):
        """Setup"""
        self.collection = RulesCollection()
        self.collection.register(Configuration())

    def tearDown(self):
        """Tear Down"""
        # Reset the Spec override to prevent other tests to fail
        cfnlint.helpers.initialize_specs()

    def test_fail_run(self):
        """Failure test required"""
        filename = "test/fixtures/templates/bad/override/include.yaml"
        template = self.load_template(filename)

        with open("test/fixtures/templates/override_spec/include.json") as fp:
            custom_spec = json.load(fp)
        cfnlint.helpers.set_specs(custom_spec)

        bad_runner = Runner(self.collection, filename, template, ["us-east-1"], [])
        errs = bad_runner.run()
        self.assertEqual(2, len(errs))



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.outputs.Value import Value  # pylint: disable=E0401


class TestOutputValue(BaseRuleTestCase):
    """Test template parameter configurations"""

    def setUp(self):
        """Setup"""
        super(TestOutputValue, self).setUp()
        self.collection.register(Value())

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/outputs/configuration.yaml", 2
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import logging
import sys
from test.testlib.testcase import BaseTestCase
from unittest.mock import call, mock_open, patch

import cfnlint.maintenance
from cfnlint.rules import CloudFormationLintRule, RulesCollection

LOGGER = logging.getLogger("cfnlint.maintenance")
LOGGER.addHandler(logging.NullHandler())


class TestUpdateDocumentation(BaseTestCase):
    """Used for Testing Updating the Documentation"""

    TEST_TEXT = """
Regular Text
## Rules
"""

    def test_update_docs(self):
        collection = RulesCollection(include_rules=["I"], include_experimental=True)

        builtin_module_name = "builtins"

        mo = mock_open(read_data=self.TEST_TEXT)
        mo.return_value.__iter__ = lambda self: self
        mo.return_value.__iter__ = lambda self: iter(self.readline, "")
        with patch("{}.open".format(builtin_module_name), mo) as mock_builtin_open:
            cfnlint.maintenance.update_documentation(collection)

            expected_calls = [
                call("\n"),
                call("Regular Text\n"),
                call("## Rules\n"),
                call(
                    "(_This documentation is generated by running `cfn-lint --update-documentation`, do not alter this manually_)\n\n"
                ),
                call(
                    "The following **{}** rules are applied by this linter:\n\n".format(
                        len(collection) + 3
                    )
                ),
                call(
                    "| Rule ID  | Title | Description | Config<br />(Name:Type:Default) | Source | Tags |\n"
                ),
                call(
                    "| -------- | ----- | ----------- | ---------- | ------ | ---- |\n"
                ),
                call(
                    '| [E0000<a name="E0000"></a>](../src/cfnlint/rules/__init__.py) | Parsing error found when parsing the template | Checks for JSON/YAML formatting errors in your template |  | [Source](https://github.com/aws-cloudformation/cfn-python-lint) | `base` |\n'
                ),
                call(
                    '| [E0001<a name="E0001"></a>](../src/cfnlint/rules/__init__.py) | Error found when transforming the template | Errors found when performing transformation on the template |  | [Source](https://github.com/aws-cloudformation/cfn-python-lint) | `base`,`transform` |\n'
                ),
                call(
                    '| [E0002<a name="E0002"></a>](../src/cfnlint/rules/__init__.py) | Error processing rule on the template | Errors found when processing a rule on the template |  | [Source](https://github.com/aws-cloudformation/cfn-python-lint) | `base`,`rule` |\n'
                ),
                call("\n\\* experimental rules\n"),
            ]
            mock_builtin_open.return_value.write.assert_has_calls(expected_calls)
            self.assertEqual(
                len(expected_calls), mock_builtin_open.return_value.write.call_count
            )



from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.functions.ToJsonString import ToJsonString


class TestRulesToJsonString(BaseRuleTestCase):
    def setUp(self):
        super(TestRulesToJsonString, self).setUp()
        self.collection.register(ToJsonString())
        self.success_templates = [
            "test/fixtures/templates/good/functions/toJsonString.yaml"
        ]

    def test_file_positive(self):
        self.helper_file_positive()

    def test_file_negative(self):
        self.helper_file_negative(
            "test/fixtures/templates/bad/functions/toJsonString.yaml", 3
        )

    def test_file_negative_missing_transform(self):
        self.helper_file_negative(
            "test/fixtures/templates/bad/functions/toJsonStringWithoutTransform.yaml", 1
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.resources.properties.AvailabilityZone import (
    AvailabilityZone,  # pylint: disable=E0401
)


class TestPropertyAvailabilityZone(BaseRuleTestCase):
    """Test Password Property Configuration"""

    def setUp(self):
        """Setup"""
        super(TestPropertyAvailabilityZone, self).setUp()
        self.collection.register(AvailabilityZone())
        self.success_templates = [
            "test/fixtures/templates/good/resources/properties/az.yaml"
        ]

    def test_file_positive(self):
        """Success test"""
        self.helper_file_positive()

    def test_file_negative(self):
        """Failure test"""
        self.helper_file_negative("test/fixtures/templates/bad/properties_az.yaml", 3)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import json
import sys
import xml.etree.ElementTree as ET
from test.testlib.testcase import BaseTestCase

import jsonschema

import cfnlint.core
import cfnlint.formatters
import cfnlint.helpers
from cfnlint.rules import Match
from cfnlint.rules.functions.SubUnneeded import SubUnneeded
from cfnlint.rules.resources.properties.ValuePrimitiveType import ValuePrimitiveType
from cfnlint.rules.resources.UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes import (
    UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes,
)


class TestFormatters(BaseTestCase):
    """Test Formatters"""

    def setUp(self) -> None:
        super().setUp()
        cfnlint.core._reset_rule_cache()
        self.filename = "test/fixtures/templates/bad/formatters.yaml"
        self.results = [
            Match(
                6,
                3,
                6,
                10,
                self.filename,
                UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes(),
                "The default action when replacing/removing a resource is to delete it. Set explicit values for UpdateReplacePolicy / DeletionPolicy on potentially stateful resource: Resources/myTable",
            ),
            Match(
                9,
                7,
                9,
                16,
                self.filename,
                SubUnneeded(),
                "Fn::Sub isn't needed because there are no variables at Resources/myTable/Properties/TableName/Fn::Sub",
            ),
            Match(
                18,
                9,
                18,
                27,
                self.filename,
                ValuePrimitiveType(),
                "Property Resources/myTable/Properties/ProvisionedThroughput/WriteCapacityUnits should be of type Integer",
            ),
        ]

    def _get_results(self, args, s_formatter):
        # Run a broken template
        (args, filenames, formatter) = cfnlint.core.get_args_filenames(args)

        results = []
        for filename in filenames:
            (template, rules, _) = cfnlint.core.get_template_rules(filename, args)

            results.extend(
                cfnlint.core.run_checks(filename, template, rules, ["us-east-1"])
            )
        # Validate Formatter class initiated
        self.assertEqual(s_formatter, formatter.__class__.__name__)
        # We need 3 errors (Information, Warning, Error)
        self.assertEqual(len(results), 3)

        return (results, args, filenames, formatter)

    def test_base_formatter(self):
        """Test base formatter"""

        (results, _, _, formatter) = self._get_results(
            [
                "--template",
                self.filename,
                "--include-checks",
                "I",
                "--configure-rule",
                "E3012:strict=true",
            ],
            "Formatter",
        )
        # Get the base format output
        a_results = formatter.print_matches(results).splitlines()

        # Check the errors
        self.assertEqual(
            a_results[0], f"{self.results[0].rule.id} {self.results[0].message}"
        )
        self.assertEqual(
            a_results[1],
            f"{self.results[0].filename}:{self.results[0].linenumber}:{self.results[0].columnnumber}",
        )
        self.assertEqual(a_results[2], "")
        self.assertEqual(
            a_results[3], f"{self.results[1].rule.id} {self.results[1].message}"
        )
        self.assertEqual(
            a_results[4],
            f"{self.results[1].filename}:{self.results[1].linenumber}:{self.results[1].columnnumber}",
        )
        self.assertEqual(a_results[5], "")
        self.assertEqual(
            a_results[6], f"{self.results[2].rule.id} {self.results[2].message}"
        )
        self.assertEqual(
            a_results[7],
            f"{self.results[2].filename}:{self.results[2].linenumber}:{self.results[2].columnnumber}",
        )

    def test_quiet_formatter(self):
        """Test quiet formatter"""

        (results, _, _, formatter) = self._get_results(
            [
                "--template",
                self.filename,
                "--format",
                "quiet",
                "--include-checks",
                "I",
                "--configure-rule",
                "E3012:strict=true",
            ],
            "QuietFormatter",
        )
        # Get the base format output
        a_results = formatter.print_matches(results).splitlines()

        for i in range(3):
            # Check the errors
            self.assertEqual(
                a_results[i],
                f"{self.results[i].rule.id}: {self.results[i].rule.shortdesc} {self.results[i].filename}:{self.results[i].linenumber}",
            )

    def test_parseable_formatter(self):
        """Test Parseable formatter"""

        (results, _, _, formatter) = self._get_results(
            [
                "--template",
                self.filename,
                "--format",
                "parseable",
                "--include-checks",
                "I",
                "--configure-rule",
                "E3012:strict=true",
            ],
            "ParseableFormatter",
        )
        # Get the base format output
        a_results = formatter.print_matches(results).splitlines()

        for i in range(3):
            # Check the errors
            self.assertEqual(
                a_results[i],
                f"{self.results[i].filename}:{self.results[i].linenumber}:{self.results[i].columnnumber}:{self.results[i].linenumberend}:{self.results[i].columnnumberend}:{self.results[i].rule.id}:{self.results[i].message}",
            )

    def test_pretty_formatter(self):
        """Test pretty formatter"""

        (results, args, filenames, formatter) = self._get_results(
            [
                "--template",
                self.filename,
                "--format",
                "pretty",
                "--include-checks",
                "I",
                "--configure-rule",
                "E3012:strict=true",
            ],
            "PrettyFormatter",
        )

        rules = None
        for filename in filenames:
            (_, rules, _) = cfnlint.core.get_template_rules(filename, args)

        # Get the base format output
        a_results = formatter.print_matches(results, rules, filenames).splitlines()

        if sys.stdout.isatty():
            # Check the errors
            self.assertEqual(a_results[0], f"\x1b[4m{self.filename}\x1b[0m")
            self.assertEqual(
                a_results[1],
                f"\x1b[0m{self.results[2].linenumber}:{self.results[2].columnnumber}:               \x1b[0m\x1b[31m{self.results[2].rule.id}     \x1b[0m{self.results[2].message}",
            )
            self.assertEqual(
                a_results[2],
                f"\x1b[0m{self.results[1].linenumber}:{self.results[1].columnnumber}:                \x1b[0m\x1b[33m{self.results[1].rule.id}     \x1b[0m{self.results[1].message}",
            )
            self.assertEqual(
                a_results[3],
                f"\x1b[0m{self.results[0].linenumber}:{self.results[0].columnnumber}:                \x1b[0m\x1b[34m{self.results[0].rule.id}     \x1b[0m{self.results[0].message}",
            )
        else:
            # Check the errors
            self.assertEqual(a_results[0], self.filename)
            self.assertEqual(
                a_results[1],
                f"{self.results[2].linenumber}:{self.results[2].columnnumber}:               {self.results[2].rule.id}     {self.results[2].message}",
            )
            self.assertEqual(
                a_results[2],
                f"{self.results[1].linenumber}:{self.results[1].columnnumber}:                {self.results[1].rule.id}     {self.results[1].message}",
            )
            self.assertEqual(
                a_results[3],
                f"{self.results[0].linenumber}:{self.results[0].columnnumber}:                {self.results[0].rule.id}     {self.results[0].message}",
            )

    def test_json_formatter(self):
        """Test JSON formatter"""
        (results, _, _, formatter) = self._get_results(
            [
                "--template",
                self.filename,
                "--format",
                "json",
                "--include-checks",
                "I",
                "--configure-rule",
                "E3012:strict=true",
            ],
            "JsonFormatter",
        )

        # Check the errors
        self.assertEqual(results[0].rule.id, "I3011")
        self.assertEqual(results[1].rule.id, "W1020")
        self.assertEqual(results[2].rule.id, "E3012")

        # Get the JSON output
        json_results = json.loads(formatter.print_matches(results))

        # Check the 3 errors again
        self.assertEqual(len(json_results), 3)

        # Check the errors level description
        for i in range(3):
            self.assertEqual(json_results[i]["Rule"]["Id"], self.results[i].rule.id)
            self.assertEqual(json_results[i]["Message"], self.results[i].message)

    def test_junit_returns_none(self):
        """Test JUnut Formatter returns None if no rules are passed in"""

        (results, _, _, formatter) = self._get_results(
            [
                "--template",
                self.filename,
                "--format",
                "junit",
                "--include-checks",
                "I",
                "--ignore-checks",
                "E1029",
                "--configure-rule",
                "E3012:strict=true",
            ],
            "JUnitFormatter",
        )

        # The actual test
        self.assertIsNone(formatter.print_matches([], []))

    def test_junit_formatter(self):
        """Test JUnit Formatter"""
        (results, _, _, formatter) = self._get_results(
            [
                "--template",
                self.filename,
                "--format",
                "junit",
                "--include-checks",
                "I",
                "--ignore-checks",
                "E1029",
                "--configure-rule",
                "E3012:strict=true",
            ],
            "JUnitFormatter",
        )

        # Check the errors
        self.assertEqual(results[0].rule.id, "I3011")
        self.assertEqual(results[1].rule.id, "W1020")
        self.assertEqual(results[2].rule.id, "E3012")

        root = ET.fromstring(
            formatter.print_matches(results, cfnlint.core.get_used_rules())
        )

        self.assertEqual(root.tag, "testsuites")
        self.assertEqual(root[0].tag, "testsuite")

        found_i3011 = False
        found_w1020 = False
        found_e3012 = False
        name_i3011 = "{0} {1}".format(results[0].rule.id, results[0].rule.shortdesc)
        name_w1020 = "{0} {1}".format(results[1].rule.id, results[1].rule.shortdesc)
        name_e3012 = "{0} {1}".format(results[2].rule.id, results[2].rule.shortdesc)
        name_e1029 = "E1029 Sub is required if a variable is used in a string"
        for child in root[0]:
            self.assertEqual(child.tag, "testcase")

            if child.attrib["name"] in [name_i3011, name_w1020, name_e3012]:
                self.assertEqual(child[0].tag, "failure")

                if child.attrib["name"] == name_i3011:
                    found_i3011 = True
                if child.attrib["name"] == name_w1020:
                    found_w1020 = True
                if child.attrib["name"] == name_e3012:
                    found_e3012 = True

            if child.attrib["name"] == name_e1029:
                self.assertEqual(child[0].tag, "skipped")
                self.assertEqual(child[0].attrib["type"], "skipped")

        self.assertTrue(found_i3011)
        self.assertTrue(found_w1020)
        self.assertTrue(found_e3012)

    def test_sarif_formatter(self):
        """Test the SARIF formatter"""

        (results, args, filenames, formatter) = self._get_results(
            [
                "--template",
                self.filename,
                "--format",
                "sarif",
                "--include-checks",
                "I",
                "--ignore-checks",
                "E1029",
                "--configure-rule",
                "E3012:strict=true",
            ],
            "SARIFFormatter",
        )

        rules = None
        for filename in filenames:
            (_, rules, _) = cfnlint.core.get_template_rules(filename, args)

        # Check the errors
        self.assertEqual(results[0].rule.id, "I3011")
        self.assertEqual(results[1].rule.id, "W1020")
        self.assertEqual(results[2].rule.id, "E3012")

        # Get the SARIF JSON output
        sarif = json.loads(formatter.print_matches(results, rules))

        # Fetch the SARIF schema
        schema = json.loads(cfnlint.helpers.get_url_content(sarif["$schema"], False))
        jsonschema.validate(sarif, schema)

        sarif_results = sarif["runs"][0]["results"]
        # Check the 3 errors again
        self.assertEqual(len(sarif_results), 3)

        # Sanity check the errors
        self.assertEqual(sarif_results[0]["level"], "note")
        self.assertEqual(sarif_results[0]["ruleId"], "I3011")
        # IMPORTANT: 'warning' is the default level in SARIF (when kind is
        # absent) and is stripped by serialization
        self.assertNotIn("level", sarif_results[1].keys())
        self.assertEqual(sarif_results[1]["ruleId"], "W1020")
        self.assertEqual(sarif_results[2]["level"], "error")
        self.assertEqual(sarif_results[2]["ruleId"], "E3012")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules.common import number_limit


class LimitNumber(CloudFormationLintRule):
    """Check if maximum Output limit is exceeded"""

    id = "E6010"
    shortdesc = "Output limit not exceeded"
    description = (
        "Check the number of Outputs in the template is less than the upper limit"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["outputs", "limits"]

    def match(self, cfn):
        return number_limit(cfn, "Outputs")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.functions.Select import Select  # pylint: disable=E0401


class TestRulesSelect(BaseRuleTestCase):
    """Test Rules Get Att"""

    def setUp(self):
        """Setup"""
        super(TestRulesSelect, self).setUp()
        self.collection.register(Select())

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/functions_select.yaml", 4
        )

    def test_select_parts(self):
        rule = Select()
        # can't be a list
        self.assertEqual(len(rule._test_index_obj([], [])), 1)
        # can't be a string
        self.assertEqual(len(rule._test_index_obj("a", [])), 1)
        # can be a valid fn
        self.assertEqual(len(rule._test_index_obj({"Ref": "Test"}, [])), 0)
        # can't be an invalid fn
        self.assertEqual(len(rule._test_index_obj({"Foo": "Bar"}, [])), 1)
        # can't be a dict of many values
        self.assertEqual(
            len(rule._test_index_obj({"Ref": "Test", "Foo": "Bar"}, [])), 1
        )

        self.assertEqual(len(rule._test_index_obj({}, [])), 1)

        # supported function
        self.assertEqual(len(rule._test_list_obj({"Ref": "Test"}, [])), 0)
        # Unsupported function
        self.assertEqual(len(rule._test_list_obj({"Foo": "Bar"}, [])), 1)
        # Unsupported type
        self.assertEqual(len(rule._test_list_obj("foo", [])), 1)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import logging
import sys
from test.testlib.testcase import BaseTestCase
from unittest.mock import ANY, MagicMock, Mock, call, patch

import cfnlint.maintenance

LOGGER = logging.getLogger("cfnlint.maintenance")
LOGGER.addHandler(logging.NullHandler())


def patch_spec_sideffect(content, region, patch_types="ExtendedSpecs"):
    if region == "all" and patch_types == "ExtendedSpecs":
        return {
            "PropertyTypes": {
                "AWS::Lambda::CodeSigningConfig.AllowedPublishers": {
                    "Properties": {
                        "SigningProfileVersionArns": {},
                    }
                },
                "AWS::Lambda::CodeSigningConfig.CodeSigningPolicies": {
                    "Properties": {
                        "UntrustedArtifactOnDeployment": {},
                    }
                },
            },
            "ResourceTypes": {
                "AWS::Lambda::CodeSigningConfig": {
                    "Attributes": {
                        "CodeSigningConfigArn": {
                            "PrimitiveType": "String",
                        },
                        "CodeSigningConfigId": {
                            "PrimitiveType": "String",
                        },
                    },
                    "Properties": {
                        "AllowedPublishers": {},
                        "CodeSigningPolicies": {},
                        "Description": {},
                    },
                }
            },
            "ValueTypes": {},
        }
    if region in ["us-east-1", "us-west-2"] and patch_types == "ExtendedSpecs":
        return {
            "PropertyTypes": {
                "AWS::Lambda::CodeSigningConfig.AllowedPublishers": {
                    "Properties": {
                        "SigningProfileVersionArns": {},
                    }
                },
                "AWS::Lambda::CodeSigningConfig.CodeSigningPolicies": {
                    "Properties": {
                        "UntrustedArtifactOnDeployment": {},
                    }
                },
            },
            "ResourceTypes": {
                "AWS::Lambda::CodeSigningConfig": {
                    "Attributes": {
                        "CodeSigningConfigArn": {
                            "PrimitiveType": "String",
                        },
                        "CodeSigningConfigId": {
                            "PrimitiveType": "String",
                        },
                    },
                    "Properties": {
                        "AllowedPublishers": {},
                        "CodeSigningPolicies": {},
                        "Description": {},
                    },
                }
            },
            "ValueTypes": {
                "AWS::EC2::Instance.Types": ["m2.medium"],
            },
        }

    return content


class TestUpdateResourceSpecs(BaseTestCase):
    """Used for Testing Resource Specs"""

    @patch("cfnlint.maintenance.url_has_newer_version")
    @patch("cfnlint.maintenance.get_url_content")
    @patch("cfnlint.maintenance.json.dump")
    @patch("cfnlint.maintenance.patch_spec")
    @patch("cfnlint.maintenance.SPEC_REGIONS", {"us-east-1": "http://foo.badurl"})
    @patch("cfnlint.maintenance.urlopen")
    def test_update_resource_spec(
        self,
        mock_urlopen,
        mock_patch_spec,
        mock_json_dump,
        mock_content,
        mock_url_newer_version,
    ):
        """Success update resource spec"""

        mock_url_newer_version.return_value = True
        mock_content.return_value = '{"PropertyTypes": {}, "ResourceTypes": {}}'
        mock_patch_spec.side_effect = patch_spec_sideffect

        cm = MagicMock()
        cm.getcode.return_value = 200
        cm.__enter__.return_value = cm

        with open("test/fixtures/registry/schema.zip", "rb") as f:
            byte = f.read()
            cm.read.return_value = byte

        mock_urlopen.return_value = cm
        schema_cache = cfnlint.maintenance.get_schema_value_types()

        builtin_module_name = "builtins"

        with patch("{}.open".format(builtin_module_name)) as mock_builtin_open:
            cfnlint.maintenance.update_resource_spec(
                "us-east-1", "http://foo.badurl", schema_cache
            )
            mock_json_dump.assert_called_with(
                {
                    "PropertyTypes": {
                        "AWS::Lambda::CodeSigningConfig.AllowedPublishers": {
                            "Properties": {
                                "SigningProfileVersionArns": {
                                    "Value": {
                                        "ValueType": "AWS::Lambda::CodeSigningConfig.AllowedPublishers.SigningProfileVersionArns",
                                    }
                                },
                            }
                        },
                        "AWS::Lambda::CodeSigningConfig.CodeSigningPolicies": {
                            "Properties": {
                                "UntrustedArtifactOnDeployment": {
                                    "Value": {
                                        "ValueType": "AWS::Lambda::CodeSigningConfig.CodeSigningPolicies.UntrustedArtifactOnDeployment",
                                    }
                                },
                            }
                        },
                    },
                    "ResourceTypes": {
                        "AWS::Lambda::CodeSigningConfig": {
                            "Attributes": {
                                "CodeSigningConfigArn": {
                                    "PrimitiveType": "String",
                                },
                                "CodeSigningConfigId": {
                                    "PrimitiveType": "String",
                                },
                            },
                            "Properties": {
                                "AllowedPublishers": {},
                                "CodeSigningPolicies": {},
                                "Description": {
                                    "Value": {
                                        "ValueType": "AWS::Lambda::CodeSigningConfig.Description"
                                    },
                                },
                            },
                        }
                    },
                    "ValueTypes": {
                        "AWS::EC2::Instance.Types": ["m2.medium"],
                        "AWS::Lambda::CodeSigningConfig.Description": {
                            "StringMin": 0,
                            "StringMax": 256,
                        },
                        "AWS::Lambda::CodeSigningConfig.AllowedPublishers.SigningProfileVersionArns": {
                            "AllowedPatternRegex": "arn:(aws[a-zA-Z0-9-]*):([a-zA-Z0-9\\-])+:([a-z]{2}(-gov)?-[a-z]+-\\d{1})?:(\\d{12})?:(.*)",
                            "StringMin": 12,
                            "StringMax": 1024,
                        },
                        "AWS::Lambda::CodeSigningConfig.CodeSigningPolicies.UntrustedArtifactOnDeployment": {
                            "AllowedValues": ["Warn", "Enforce"]
                        },
                    },
                },
                mock_builtin_open.return_value.__enter__.return_value,
                indent=1,
                separators=(",", ": "),
                sort_keys=True,
            )

    @patch("cfnlint.maintenance.url_has_newer_version")
    @patch("cfnlint.maintenance.get_url_content")
    @patch("cfnlint.maintenance.json.dump")
    @patch("cfnlint.maintenance.patch_spec")
    @patch("cfnlint.maintenance.SPEC_REGIONS", {"us-east-1": "http://foo.badurl"})
    @patch("cfnlint.maintenance.urlopen")
    @patch("cfnlint.maintenance.cfnlint.helpers.load_resource")
    def test_update_resource_spec_cache(
        self,
        mock_load_resource,
        mock_urlopen,
        mock_patch_spec,
        mock_json_dump,
        mock_content,
        mock_url_newer_version,
    ):
        """Success update resource spec with cache"""

        mock_url_newer_version.return_value = True
        mock_content.return_value = '{"PropertyTypes": {}, "ResourceTypes": {}}'
        mock_patch_spec.side_effect = patch_spec_sideffect

        cm = MagicMock()
        cm.getcode.return_value = 200
        cm.__enter__.return_value = cm

        with open("test/fixtures/registry/schema.zip", "rb") as f:
            byte = f.read()
            cm.read.return_value = byte

        mock_urlopen.return_value = cm
        schema_cache = cfnlint.maintenance.get_schema_value_types()

        builtin_module_name = "builtins"

        mock_load_resource.return_value = {
            "PropertyTypes": {
                "AWS::Lambda::CodeSigningConfig.AllowedPublishers": {
                    "Properties": {
                        "SigningProfileVersionArns": {
                            "Value": {
                                "ValueType": "AWS::Lambda::CodeSigningConfig.AllowedPublishers.SigningProfileVersionArns",
                            }
                        },
                    }
                },
                "AWS::Lambda::CodeSigningConfig.CodeSigningPolicies": {
                    "Properties": {
                        "UntrustedArtifactOnDeployment": {
                            "Value": {
                                "ValueType": "AWS::Lambda::CodeSigningConfig.CodeSigningPolicies.UntrustedArtifactOnDeployment",
                            }
                        },
                    }
                },
            },
            "ResourceTypes": {
                "AWS::Lambda::CodeSigningConfig": {
                    "Attributes": {
                        "CodeSigningConfigArn": {
                            "PrimitiveType": "String",
                        },
                        "CodeSigningConfigId": {
                            "PrimitiveType": "String",
                        },
                    },
                    "Properties": {
                        "AllowedPublishers": {},
                        "CodeSigningPolicies": {},
                        "Description": {
                            "Value": {
                                "ValueType": "AWS::Lambda::CodeSigningConfig.Description"
                            },
                        },
                    },
                }
            },
            "ValueTypes": {
                "AWS::EC2::Instance.Types": [
                    "m2.medium",
                    "m2.large",
                ],
                "AWS::Lambda::CodeSigningConfig.Description": {
                    "StringMin": 0,
                    "StringMax": 256,
                },
                "AWS::Lambda::CodeSigningConfig.AllowedPublishers.SigningProfileVersionArns": {
                    "AllowedPatternRegex": "arn:(aws[a-zA-Z0-9-]*):([a-zA-Z0-9\\-])+:([a-z]{2}(-gov)?-[a-z]+-\\d{1})?:(\\d{12})?:(.*)",
                    "StringMin": 12,
                    "StringMax": 1024,
                },
                "AWS::Lambda::CodeSigningConfig.CodeSigningPolicies.UntrustedArtifactOnDeployment": {
                    "AllowedValues": ["Warn", "Enforce"]
                },
            },
        }

        with patch("{}.open".format(builtin_module_name)) as mock_builtin_open:
            cfnlint.maintenance.update_resource_spec(
                "us-west-2", "http://foo.badurl", schema_cache
            )
            mock_json_dump.assert_called_with(
                {
                    "PropertyTypes": {
                        "AWS::Lambda::CodeSigningConfig.AllowedPublishers": "CACHED",
                        "AWS::Lambda::CodeSigningConfig.CodeSigningPolicies": "CACHED",
                    },
                    "ResourceTypes": {
                        "AWS::Lambda::CodeSigningConfig": "CACHED",
                    },
                    "ValueTypes": {
                        # does not match above
                        "AWS::EC2::Instance.Types": [
                            "m2.medium",
                        ],
                        "AWS::Lambda::CodeSigningConfig.Description": "CACHED",
                        "AWS::Lambda::CodeSigningConfig.AllowedPublishers.SigningProfileVersionArns": "CACHED",
                        "AWS::Lambda::CodeSigningConfig.CodeSigningPolicies.UntrustedArtifactOnDeployment": "CACHED",
                    },
                },
                mock_builtin_open.return_value.__enter__.return_value,
                indent=1,
                separators=(",", ": "),
                sort_keys=True,
            )

    @patch("cfnlint.maintenance.url_has_newer_version")
    @patch("cfnlint.maintenance.get_url_content")
    @patch("cfnlint.maintenance.json.dump")
    @patch("cfnlint.maintenance.patch_spec")
    @patch("cfnlint.maintenance.SPEC_REGIONS", {"us-east-1": "http://foo.badurl"})
    def test_do_not_update_resource_spec(
        self, mock_patch_spec, mock_json_dump, mock_content, mock_url_newer_version
    ):
        """Success update resource spec"""

        mock_url_newer_version.return_value = False

        result = cfnlint.maintenance.update_resource_spec(
            "us-east-1", "http://foo.badurl", ANY
        )
        self.assertIsNone(result)
        mock_content.assert_not_called()
        mock_patch_spec.assert_not_called()
        mock_json_dump.assert_not_called()

    @patch("cfnlint.maintenance.url_has_newer_version")
    @patch("cfnlint.maintenance.get_url_content")
    @patch("cfnlint.maintenance.json.dump")
    @patch("cfnlint.maintenance.patch_spec")
    @patch("cfnlint.maintenance.SPEC_REGIONS", {"us-east-1": "http://foo.badurl"})
    @patch("cfnlint.maintenance.urlopen")
    def test_update_resource_spec_force(
        self,
        mock_urlopen,
        mock_patch_spec,
        mock_json_dump,
        mock_content,
        mock_url_newer_version,
    ):
        """Success update resource spec"""

        mock_url_newer_version.return_value = False
        mock_content.return_value = '{"PropertyTypes": {}, "ResourceTypes": {}}'
        mock_patch_spec.side_effect = patch_spec_sideffect

        mock_urlresponse = Mock()
        cm = MagicMock()
        cm.getcode.return_value = 200
        cm.__enter__.return_value = cm

        with open("test/fixtures/registry/schema.zip", "rb") as f:
            byte = f.read()
            cm.read.return_value = byte

        mock_urlopen.return_value = cm
        schema_cache = cfnlint.maintenance.get_schema_value_types()

        builtin_module_name = "builtins"

        with patch("{}.open".format(builtin_module_name)) as mock_builtin_open:
            cfnlint.maintenance.update_resource_spec(
                "us-east-1", "http://foo.badurl", schema_cache, True
            )
            mock_content.assert_called_once()
            mock_patch_spec.assert_called()
            mock_json_dump.assert_called_once()

    @patch("cfnlint.maintenance.multiprocessing.Pool")
    @patch("cfnlint.maintenance.update_resource_spec")
    @patch("cfnlint.maintenance.SPEC_REGIONS", {"us-east-1": "http://foo.badurl"})
    def test_update_resource_specs_python_3(self, mock_update_resource_spec, mock_pool):
        fake_pool = MagicMock()
        mock_pool.return_value.__enter__.return_value = fake_pool

        cfnlint.maintenance.update_resource_specs()

        fake_pool.starmap.assert_called_once()
